"""xautic: live reloading with Python."""

from _xautic import __version__
from _xautic.main import StatReloader
from _xautic.main import debug
from _xautic.main import restart_with_reloader
from _xautic.main import run_with_reloader

__all__ = [
    "__version__",
    "StatReloader",
    "debug",
    "restart_with_reloader",
    "run_with_reloader",
]



"""
Changelog and version history for xautic.

.. versionadded:: 1.0.0
    Built and setup the structure for the application.

.. versionadded:: 1.0.0
    Support for `debug` decorator was added along with major code
    refactoring.
"""

__all__ = ["__version__"]

__version__ = "1.0.0"



"""
Useful utilities.

Utility functions that offers general convenience to facilitate live
reloading and tracking of the necessary components. Primarily these
functions are used within xautic, but they can also be used for external
consumption.
"""

import fnmatch
import itertools
import logging
import os
import sys
from typing import Any
from typing import Iterator
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union
from typing import cast

PathLike = Union[str, os.PathLike]

_logger: Optional[logging.Logger] = None

# These are the directories that we purposely ignore while tracking the
# changes within the working root. This optimization reduces the CPU
# utilization significantly since there will be few files to track.
IGNORED_DIRS: Set[PathLike] = {
    ".egg-info",
    ".git",
    ".hg",
    ".mypy_cache",
    ".pytest_cache",
    "__pycache__",
}
VALID_PY_FILES: Tuple[str, ...] = (".py", ".pyc", ".pyw")

# All possible system prefixes from where the imports can be found. The
# base values are different when running from a virtualenv. The reloader
# won't scan these directories as it would be too inefficient.
prefixes: Union[Set[PathLike], Tuple[PathLike, ...]] = {
    sys.prefix,
    sys.base_prefix,
    sys.exec_prefix,
    sys.base_exec_prefix,
}

if hasattr(sys, "real_prefix"):
    prefixes.add(sys.real_prefix) # type: ignore[attr-defined,union-attr]

prefixes = tuple(prefixes)


def _has_level_handler(logger: logging.Logger) -> bool:
    """
    Check if there is a handler in the logging chain that will handle
    the given logger's effective level.

    :param logger: Logger object.
    :return: Handler level state.
    """
    level = logger.getEffectiveLevel()
    current = logger
    while current:
        if any(handler.level <= level for handler in current.handlers):
            return True
        if not current.propagate:
            break
        current = current.parent  # type: ignore[assignment]
    return False


def _log(level: str, msg: str, *args: Any, **kwargs: Any) -> None:
    """
    Log messages to the `xautic` logger.

    The logger is created the first time it is needed. If there is no
    level set, it is set to :py:data:`logging.INFO`. If there is no
    handler added for the logger's effective level,
    :py:class:`~logging.StreamHandler` is added.

    :param level: Logging level severity.
    :param msg: Message to be logged.

    .. warning::
        This is not a public API anymore, do not use this function.
    """
    global _logger
    if _logger is None:
        _logger = logging.getLogger("xautic.main")
        if _logger.level == logging.NOTSET:
            _logger.setLevel(logging.INFO)
        if not _has_level_handler(_logger):
            _logger.addHandler(logging.StreamHandler())
    getattr(_logger, level)(msg.rstrip(), *args, **kwargs)


def _imported_module_paths() -> Iterator[PathLike]:
    """
    Yield absolute paths of the imported modules.

    Loop over all the imported modules and try to get their absolute
    file paths. If the path consists of all directories and no file,
    then the implementation will break the loop and move on.
    """
    for module in list(sys.modules.values()):
        path = getattr(module, "__file__", None)
        if path is None:
            continue
        # We check if the module is running from a file or a directory.
        # If the path is a directory, we skip traversing the loop.
        while not os.path.isfile(path):
            tmp = path
            path = os.path.dirname(path)
            if path == tmp:
                break
        else:
            yield path


def _all_possible_paths(
    track: Set[PathLike], ignore_patterns: Set[str]
) -> Set[PathLike]:
    """
    Return list of paths for the reloader to track.

    Paths can be both for python and non-python files. While tracking
    python files, it also tracks the absolute paths of the imported
    modules. Here, we purposely exclude the system paths for efficiency.

    :param track: Absolute paths of both python & non-python files to
        track. Well, you can add non-python files to track but doing
        so might increase the CPU utilization. It is more advisable to
        track only python files e.g.: .py or .pyc.
    :param ignore_patterns: Path patterns to ignore. The patterns can be
        regular expressions or just names of the file(s) and directories
        to be ignored.
    :return: Set of all paths to track including the imported module
        paths and non-system file paths (python and non-python files).
    """
    paths = set()
    for path in itertools.chain(list(sys.path), track):
        path = os.path.abspath(path)
        if os.path.isfile(path):
            paths.add(path)
        for root, dirs, files in os.walk(path):
            if (
                root.startswith(prefixes)
                or os.path.basename(root) in IGNORED_DIRS
            ):
                dirs.clear()
                continue
            for file in files:
                if file.endswith(VALID_PY_FILES):
                    paths.add(os.path.join(root, file))
    paths.update(_imported_module_paths())
    for pattern in ignore_patterns:
        paths.difference_update(fnmatch.filter(paths, pattern))
    return paths


def _get_args_for_reloading() -> List[str]:
    """Return executable args."""
    execs = [sys.executable]
    py_script, *args = sys.argv
    # Here we're trying to understand how the script was called. There
    # are a lot of moving parts involved in this process. Simply running
    # `python3 test.py` might not be the case all the time.
    __main__ = sys.modules["__main__"]
    if getattr(__main__, "__package__", None) is None or (
        os.name == "nt"
        and __main__.__package__ == ""
        and not os.path.exists(py_script)
        and os.path.exists(f"{py_script}.exe")
    ):
        py_script = os.path.abspath(py_script)
        if os.name == "nt":
            if not os.path.exists(py_script) and os.path.exists(
                f"{py_script}.exe"
            ):
                py_script += ".exe"
            if (
                os.path.splitext(sys.executable)[1] == ".exe"
                and os.path.splitext(py_script)[1] == ".exe"
            ):
                execs.pop(0)
        execs.append(py_script)
    else:
        if sys.argv[0] == "-m":
            args = sys.argv
        else:
            if os.path.isfile(py_script):
                py_module = cast(str, __main__.__package__)
                name = os.path.splitext(os.path.basename(py_script))[0]
                if name != "__main__":
                    py_module += f".{name}"
            else:
                py_module = py_script
            execs.extend(("-m", py_module.lstrip(".")))
    execs.extend(args)
    return execs



"""
Reloading application.

It allows live reloading of a python script. The implementation is based
on the live reloading functionality of web frameworks like Flask and
Django. Once it detects any change in the repository or the project root
files, it reloads the script.
"""

import functools
import itertools
import os
import signal
import subprocess
import sys
import threading
import time
from typing import Any
from typing import Callable
from typing import Dict
from typing import Final
from typing import Iterable
from typing import Mapping
from typing import NoReturn
from typing import Optional
from typing import Union

from ._version import __version__
from .utils import IGNORED_DIRS
from .utils import PathLike
from .utils import _all_possible_paths
from .utils import _get_args_for_reloading
from .utils import _log

try:
    import termios
except ImportError:
    termios = None  # type: ignore[assignment]

Args = Iterable
Kwargs = Mapping[str, Any]
Function = Callable[..., Any]

THREADNAME: Final[str] = "xautic-main-thread"
ENV_VAR: Final[str] = "XAUTIC_ENV"

threading.current_thread().name = THREADNAME


def _ensure_echo_on() -> None:
    """
    Ensure that echo mode is enabled. Some tools such as PDB disable it
    which causes usability issues after a reload.
    """
    if not termios or not sys.stdin.isatty():
        return
    attrs = termios.tcgetattr(sys.stdin)
    if not attrs[3] & termios.ECHO:
        attrs[3] |= termios.ECHO
        termios.tcsetattr(sys.stdin, termios.TCSANOW, attrs)


def _trigger_reload(path: PathLike) -> None:
    """Display reloading message and exit."""
    path = os.path.abspath(path)
    _log("info", f"Changes detected in {path}, refreshing script")
    sys.exit(3)


class StatReloader:
    """
    Stat based reloader.

    The reloading is triggered based on the difference of the modified
    status times of the watched files. For this, we need to know how
    many files we need to track as well as how often we should track
    them for changes. Once we have this information we can then
    calculate the modified time and later trigger the reloading.

    .. code-block:: python

        reloader = StatReloader()
        with reloader:
           reloader.run()

    .. code-block:: python

        reloader = StatReloader(interval=5.0)
        reloader.run()

    .. note::
        The current implementation is based on the stat based changes
        detected by the :py:func:`~os.stat`. This can be further
        improved using the `watchdog.observer`. But it seems that it
        increases the overhead. Whereas this implementation is minimal
        and roughly works the same.

    .. versionchanged:: 1.0.0
        In previous development version "track", "ignore_patterns" and
        "ignore_dirs" were used as means of named arguments for
        controlling the behavior of the :py:class:`StatReloader`. Now
        added support for kwargs instead.
    """

    def __init__(self, **kwargs) -> None:
        """Initialize StatReloader class with no options."""
        self.mtimes: Dict[PathLike, float] = {}
        self.track = {os.path.abspath(x) for x in kwargs.get("track") or ()}
        self.ignore_patterns = set(kwargs.get("ignore_patterns") or ())
        ignore_dirs = kwargs.get("ignore_dirs")
        if ignore_dirs:
            IGNORED_DIRS.update(ignore_dirs)
        self.interval = kwargs.get("interval", 1.0)

    def __enter__(self) -> "StatReloader":
        """
        Enter the runtime context related to this object and populate
        the initial filesystem state.
        """
        self.step_through()
        return self

    def __exit__(self, *args: Any) -> None:
        """Exit the runtime context related to this object."""

    def step_through(self) -> None:
        """
        Step through while watching the filesystem and carry on with the
        re-execution.
        """
        for path in itertools.chain(
            _all_possible_paths(self.track, self.ignore_patterns)
        ):
            try:
                curr_mtime = os.stat(path).st_mtime
            except OSError:
                continue
            prev_mtime = self.mtimes.get(path)
            if prev_mtime is None:
                self.mtimes[path] = curr_mtime
                continue
            if curr_mtime > prev_mtime:
                _trigger_reload(path)

    def run(self) -> None:
        """Continously step through and sleep after each step."""
        while 1:
            self.step_through()
            time.sleep(self.interval)


def start_xautic(
    reloader: StatReloader, func: Function, *args: Args, **kwargs: Kwargs
) -> None:
    """Start the live reloading thread."""
    _ensure_echo_on()
    xautic_main_thread = threading.Thread(
        target=func, args=args, kwargs=kwargs, name=THREADNAME
    )
    xautic_main_thread.daemon = True
    # Enter the reloader to set up an initial state, then start the main
    # thread and let the reloader update loop.
    with reloader:
        xautic_main_thread.start()
        reloader.run()


def restart_with_reloader() -> Union[NoReturn, int]:
    """
    Restart the execution in a new Python interpreter with same
    arguments.
    """
    args = _get_args_for_reloading()
    if not os.getenv(ENV_VAR):
        _log(
            "info",
            "No debugging environment found, "
            f"setting up a new environment: {ENV_VAR}",
        )
    new_environ = {**os.environ, ENV_VAR: "debug"}
    while 1:
        exit_code = subprocess.call(args, env=new_environ, close_fds=False)
        if exit_code != 3:
            return exit_code


def run_with_reloader(
    func: Optional[Function],
    *args: Args,
    track: Iterable[PathLike] = None,
    ignore_patterns: Iterable[str] = None,
    ignore_dirs: Iterable[PathLike] = None,
    interval: Union[float, int] = 1.0,
    **kwargs: Kwargs,
) -> None:
    """Run the function in an independent Python interpreter."""
    if not func:
        return
    signal.signal(signal.SIGTERM, lambda *args: sys.exit(0))
    now = time.strftime("%a %B %d, %Y - %X", time.localtime())
    script = func.__code__.co_filename
    try:
        if os.environ.get(ENV_VAR) == "debug":
            # Adding named arguments instead of kwargs to avoid conflict
            # when using or parsing the kwargs from the target function.
            reloader = StatReloader(
                track=track,
                ignore_patterns=ignore_patterns,
                ignore_dirs=ignore_dirs,
                interval=interval,
            )
            start_xautic(reloader, func, *args, **kwargs)
        else:
            _log(
                "info",
                f"Starting xautic v{__version__} live reloading "
                f"({script}) on {now}, press CTRL+C to quit",
            )
            sys.exit(restart_with_reloader())
    except KeyboardInterrupt:
        pass


def debug(
    func: Optional[Function] = None,
    track: Iterable[PathLike] = None,
    ignore_patterns: Iterable[str] = None,
    ignore_dirs: Iterable[PathLike] = None,
    interval: Union[float, int] = 1.0,
) -> Function:
    """
    Live debugging decorator.

    Decorate function that needs to be reloaded on change. This function
    internally calls :py:func:`run_with_reloader`.

    :param func: Function to run with live reloading.
    :param track: Iterable of absolute paths of both python & non-python
        files to track. Defaults to None.
    :param ignore_patterns: Iterable of path patterns to ignore. The
        patterns can be regular expressions or just names of the file(s)
        and directories to be ignored. Defaults to None.
    :param ignore_dirs: Directories to skip while watching. This does
        not expect the full absolute path of the directories, only the
        directory names will do. Defaults to None.
    :param interval: Seconds to sleep between reloading. The less the
        duration, more aggressively it will track for changes. Defaults
        to 1.0 sec.

    .. code-block:: python

        @debug
        def func(*args, **kwargs):
            ...

    .. code-block:: python

        @debug(track=["/home/.bashrc", "/home/.bash_profile"])
        def func(*args, **kwargs):
            ...

    .. note::
        Well, you can add non-python files to track but doing so might
        increase the CPU utilization. It is more advisable to track only
        python files e.g. py, pyc or pyw. 
    """
    if func is None:
        return functools.partial(
            debug,
            track=track,
            ignore_patterns=ignore_patterns,
            ignore_dirs=ignore_dirs,
            interval=interval,
        )

    @functools.wraps(func)
    def inner(*args: Args, **kwargs: Kwargs) -> None:
        return run_with_reloader(
            func,
            *args,
            track=track,
            ignore_patterns=ignore_patterns,
            ignore_dirs=ignore_dirs,
            interval=interval,
            **kwargs,
        )

    return inner



"""
xautic setup.

This will install the ``xautic`` package in the python 3.6+ environment.
Before proceeding, please ensure you have a virtualenv setup & running.

See https://github.com/xames3/xautic/ for more help.
"""

try:
    import setuptools
except ImportError:
    raise RuntimeError(
        "Could not install package in the environment as setuptools is "
        "missing. Please create a new virtual environment before proceeding."
    )

import platform

MIN_PYTHON_VERSION = ("3", "6")

if platform.python_version_tuple() < MIN_PYTHON_VERSION:
    raise SystemExit(
        "Could not install xautic in the environment. It requires python "
        f"version 3.6+, you are using {platform.python_version()}"
    )

if __name__ == "__main__":
    setuptools.setup()
